﻿Imports System.Runtime.CompilerServices
Public Module Feiertage

    Public Enum Bundesland
        'Quelle Statistisches Bundesamt: https://www.destatis.de/DE/Themen/Laender-Regionen/Regionales/Gemeindeverzeichnis/Glossar/bundeslaender.html
        None = 0
        SH = 1 '01 Schleswig-Holstein (SH)
        HH = 2 '02 Hamburg (HH)
        NI = 3 '03 Niedersachsen (NI)
        HB = 4 '04 Bremen (HB)
        NW = 5 '05 Nordrhein-Westfalen (NW)
        HE = 6 '06 Hessen (HE)
        RP = 7 '07 Rheinland-Pfalz (RP)
        BW = 8 '08 Baden-Württemberg (BW)
        BY = 9 '09 Bayern (BY)
        SL = 10 '10 Saarland (SL)
        BE = 11 '11 Berlin (BE)
        BB = 12 '12 Brandenburg (BB)
        MV = 13 '13 Mecklenburg-Vorpommern (MV)
        SN = 14 '14 Sachsen (SN)
        ST = 15 '15 Sachsen-Anhalt (ST)
        TH = 16 '16 Thüringen (TH)
    End Enum

    Public Enum Feiertag
        'Quelle: https://www.deutschland-feiert.de/feiertage/ und https://www.schulferien.org/deutschland/feiertage/2021/
        KeinFeiertag = 0
        Neujahr = 1 + (1 << 16)                       'FEST (01.01.) -> alle BL
        HeiligeDreiKönige = 6 + (1 << 16)             'FEST (06.01.) -> BW,BY,ST
        'InternationalerFrauentag                     ' -> BE
        TagDerArbeit = 1 + (5 << 16)                  'FEST (01.05.) -> alle BL
        MariäHimmelfahrt = 15 + (8 << 16)             'FEST (15.08.) -> BY,SL
        'Weltkindertag                                ' -> TH
        TagDerDeutschenEinheit = 3 + (10 << 16)       'FEST (03.10.) -> alle BL
        Reformationstag = 31 + (10 << 16)             'FEST (31.10.) -> BB,HB,HH,MV,NI,SN,ST,SH,TH
        Allerheiligen = 1 + (11 << 16)                'FEST (01.11.) -> BW,BY,NW,RP,SL
        ErsterWeihnachtsfeiertag = 25 + (12 << 16)    'FEST (25.12.) -> alle BL
        ZweiterWeihnachtsfeiertag = 26 + (12 << 16)   'FEST (26.12.) -> alle BL
        'Bewegliche Feiertage müssen in der Reihenfolge NACH den Festen stehen, da ansonsten nach
        'Abfrage eines festen Feiertags die beweglichen nur hochgezählt werden und landen so auf dem 
        'falschen Datum
        Karfreitag = 2                                'BEWEGLICH -> alle BL
        Ostersonntag                                  'BEWEGLICH -> BB
        Ostermontag                                   'BEWEGLICH -> alle BL
        ChristiHimmelfahrt                            'BEWEGLICH -> alle BL
        PfingstSonntag                                'BEWEGLICH -> BB
        PfingstMontag                                 'BEWEGLICH -> alle BL
        Fronleichnam                                  'BEWEGLICH -> BW,BY,HE,NW,RP,SL
        BussUndBettag                                 'BEWEGLICH -> SN
    End Enum

    Private _Feiertag As New HashSet(Of Integer)(DirectCast([Enum].GetValues(GetType(Feiertag)), Integer()))
    <Extension()>
    Public Function GetFeiertag(ByVal dtDate As Date) As Feiertag
        dtDate = dtDate.Date
        Dim n = dtDate.Day + (dtDate.Month << 16)
        If _Feiertag.Contains(n) Then Return DirectCast(n, Feiertag)
        Dim Ostersonntag = GetOstersonntag(dtDate.Year)
        Dim BussUndBettag = GetBussUndBettag(dtDate.Year)
        If dtDate = Ostersonntag.AddDays(-2) Then Return Feiertag.Karfreitag
        If dtDate = Ostersonntag.AddDays(1) Then Return Feiertag.Ostermontag
        If dtDate = Ostersonntag Then Return Feiertag.Ostersonntag
        If dtDate = Ostersonntag.AddDays(39) Then Return Feiertag.ChristiHimmelfahrt
        If dtDate = Ostersonntag.AddDays(49) Then Return Feiertag.PfingstSonntag
        If dtDate = Ostersonntag.AddDays(50) Then Return Feiertag.PfingstMontag
        If dtDate = Ostersonntag.AddDays(60) Then Return Feiertag.Fronleichnam
        If dtDate = BussUndBettag Then Return Feiertag.BussUndBettag
        Return Feiertag.KeinFeiertag
    End Function

    Private Function GetOstersonntag(ByVal year As Integer) As Date
        Dim a As Integer = year Mod 19
        Dim b As Integer = year \ 100
        Dim c As Integer = (8 * b + 13) \ 25 - 2
        Dim d As Integer = b - (year \ 400) - 2
        Dim e As Integer = (19 * (year Mod 19) + ((15 - c + d) Mod 30)) Mod 30
        If e = 28 Then
            If a > 10 Then
                e = 27
            End If
        ElseIf e = 29 Then
            e = 28
        End If
        Dim f As Integer = (d + 6 * e + 2 * (year Mod 4) + 4 * (year Mod 7) + 6) Mod 7
        Return New DateTime(year, 3, 22).AddDays(e + f)
    End Function

    Private Function GetBussUndBettag(ByVal year As Integer) As Date
        Dim datum = New DateTime(year, 12, 25)
        Dim day As Integer = CInt(datum.DayOfWeek)
        day = If((day = 0), -11, -day - 4)
        datum = datum.AddDays(-4 * 7).AddDays(day)
        Return datum
    End Function

    Public Function Bundeslandname(ByVal bl As Bundesland) As String
        Dim blName = "kein Bundesland gefunden"
        Select Case bl
            Case Bundesland.SH : blName = "Schlewsig-Holstein"
            Case Bundesland.HH : blName = "Hamburg"
            Case Bundesland.NI : blName = "Niedersachsen"
            Case Bundesland.HB : blName = "Bremen"
            Case Bundesland.NW : blName = "Nordrhein-Westfalen"
            Case Bundesland.HE : blName = "Hessen"
            Case Bundesland.RP : blName = "Rheinland-Pfalz"
            Case Bundesland.BW : blName = "Baden-Württemberg"
            Case Bundesland.BY : blName = "Bayern"
            Case Bundesland.SL : blName = "Saarland"
            Case Bundesland.BE : blName = "Berlin"
            Case Bundesland.BB : blName = "Brandenburg"
            Case Bundesland.MV : blName = "Mecklenburg-Vorpommern"
            Case Bundesland.SN : blName = "Sachsen"
            Case Bundesland.ST : blName = "Sachsen-Anhalt"
            Case Bundesland.TH : blName = "Thüringen"
        End Select
        Return blName
    End Function

    Public Function IstFeiertagIn(ByVal ft As Feiertag, Optional ByVal bl As Bundesland = 0) As Boolean
        Dim bool = False
        Select Case ft
            Case Feiertag.Neujahr, Feiertag.TagDerArbeit, Feiertag.TagDerDeutschenEinheit, Feiertag.ErsterWeihnachtsfeiertag,
                 Feiertag.ZweiterWeihnachtsfeiertag, Feiertag.Karfreitag, Feiertag.Ostermontag, Feiertag.ChristiHimmelfahrt, Feiertag.PfingstMontag
                bool = False
            Case Feiertag.HeiligeDreiKönige
                bool = (bl And (Bundesland.BW Or Bundesland.BY Or Bundesland.ST)) = bl
            Case Feiertag.Fronleichnam
                bool = (bl And (Bundesland.BW Or Bundesland.BY Or Bundesland.HE Or Bundesland.NW Or Bundesland.RP Or Bundesland.SL)) = bl
            Case Feiertag.MariäHimmelfahrt
                bool = (bl And (Bundesland.BY Or Bundesland.SL)) = bl
            Case Feiertag.Reformationstag
                bool = (bl And (Bundesland.BB Or Bundesland.HB Or Bundesland.HH Or Bundesland.MV Or Bundesland.NI Or Bundesland.SN Or Bundesland.ST Or Bundesland.SH Or Bundesland.TH)) = bl
            Case Feiertag.Allerheiligen
                bool = (bl And (Bundesland.BW Or Bundesland.BY Or Bundesland.NW Or Bundesland.RP Or Bundesland.SL)) = bl
            Case Feiertag.Ostersonntag, Feiertag.PfingstSonntag
                bool = (bl And (Bundesland.BB)) = bl
            Case Feiertag.BussUndBettag
                bool = (bl And (Bundesland.SN)) = bl
        End Select
        Return bool
    End Function
End Module
